@@ -29,6 +29,7 @@
 #include <unistd.h>
 #include <assert.h>
 
+#define INCLUDED_BY_MAIN_MODULE 1
 #include "../common/i18n.h"
 #include "../common/util.h"
 #include "../common/asshelp.h"
@@ -39,6 +40,7 @@
 #  include "../common/exechelp.h"
 #endif
 #include "../common/init.h"
+#include "../common/comopt.h"
 
 
 #define CONTROL_D ('D' - 'A' + 1)
@@ -58,6 +60,7 @@ enum cmd_and_opt_values
     oExec       = 'E',
     oRun        = 'r',
     oSubst      = 's',
+    oUnBuffered = 'u',
 
     oNoVerbose	= 500,
     oHomedir,
@@ -110,6 +113,7 @@ static gpgrt_opt_t opts[] = {
   ARGPARSE_s_s (oDirmngrProgram, "dirmngr-program", "@"),
   ARGPARSE_s_s (oKeyboxdProgram, "keyboxd-program", "@"),
   ARGPARSE_s_s (oChUid,          "chuid",           "@"),
+  ARGPARSE_s_n (oUnBuffered,     "unbuffered", "@"),
 
   ARGPARSE_end ()
 };
@@ -137,6 +141,7 @@ struct
   int enable_varsubst;  /* Set if variable substitution is enabled.  */
   int trim_leading_spaces;
   int no_history;
+  int unbuffered; /* Set if unbuffered mode for stdin/out is preferred.  */
 } opt;
 
 
@@ -1005,10 +1010,7 @@ do_open (char *line)
   if (fd >= 0 && fd < DIM (open_fd_table))
     {
       open_fd_table[fd].inuse = 1;
-#ifdef HAVE_W32CE_SYSTEM
-# warning fixme: implement our pipe emulation.
-#endif
-#if defined(HAVE_W32_SYSTEM) && !defined(HAVE_W32CE_SYSTEM)
+#if defined(HAVE_W32_SYSTEM)
       {
         HANDLE prochandle, handle, newhandle;
 
@@ -1038,12 +1040,12 @@ do_open (char *line)
         log_info ("file '%s' opened in \"%s\" mode, fd=%d  (libc=%d)\n",
                    name, mode, (int)open_fd_table[fd].handle, fd);
       set_int_var (varname, (int)open_fd_table[fd].handle);
-#else
+#else /* Unix */
       if (opt.verbose)
         log_info ("file '%s' opened in \"%s\" mode, fd=%d\n",
                    name, mode, fd);
       set_int_var (varname, fd);
-#endif
+#endif /* Unix */
     }
   else
     {
@@ -1242,6 +1244,7 @@ main (int argc, char **argv)
           opt.trim_leading_spaces = 1;
           break;
         case oChUid:     changeuser = pargs.r.ret_str; break;
+        case oUnBuffered: opt.unbuffered = 1; break;
 
         default: pargs.err = 2; break;
 	}
@@ -1254,6 +1257,16 @@ main (int argc, char **argv)
   if (log_get_errorcount (0))
     exit (2);
 
+  /* Process common component options.  Note that we set the config
+   * dir only here so that --homedir will have an effect.  */
+  gpgrt_set_confdir (GPGRT_CONFDIR_SYS, gnupg_sysconfdir ());
+  gpgrt_set_confdir (GPGRT_CONFDIR_USER, gnupg_homedir ());
+  if (parse_comopt (GNUPG_MODULE_NAME_CONNECT_AGENT, opt.verbose > 1))
+    exit(2);
+
+  if (comopt.no_autostart)
+     opt.autostart = 0;
+
   /* --uiserver is a shortcut for a specific raw socket.  This comes
        in particular handy on Windows. */
   if (opt.use_uiserver)
@@ -1318,8 +1331,7 @@ main (int argc, char **argv)
       assuan_fd_t no_close[3];
 
       no_close[0] = assuan_fd_from_posix_fd (es_fileno (es_stderr));
-      no_close[1] = assuan_fd_from_posix_fd (log_get_fd ());
-      no_close[2] = ASSUAN_INVALID_FD;
+      no_close[1] = ASSUAN_INVALID_FD;
 
       rc = assuan_new (&ctx);
       if (rc)
@@ -1402,6 +1414,11 @@ main (int argc, char **argv)
         log_info (_("receiving line failed: %s\n"), gpg_strerror (rc) );
     }
 
+  if (!script_fp && opt.unbuffered)
+    {
+      gpgrt_setvbuf (gpgrt_stdin, NULL, _IONBF, 0);
+      setvbuf (stdout, NULL, _IONBF, 0);
+    }
 
   for (loopidx=0; loopidx < DIM (loopstack); loopidx++)
     loopstack[loopidx].collecting = 0;
@@ -1969,6 +1986,7 @@ handle_inquire (assuan_context_t ctx, char *line)
   FILE *fp = NULL;
   char buffer[1024];
   int rc, n;
+  int cancelled = 0;
 
   /* Skip the command and trailing spaces. */
   for (; *line && !spacep (line); line++)
@@ -2013,11 +2031,7 @@ handle_inquire (assuan_context_t ctx, char *line)
     {
       if (d->is_prog)
         {
-#ifdef HAVE_W32CE_SYSTEM
-          fp = NULL;
-#else
           fp = popen (d->file, "r");
-#endif
           if (!fp)
             log_error ("error executing '%s': %s\n",
                        d->file, strerror (errno));
@@ -2050,21 +2064,20 @@ handle_inquire (assuan_context_t ctx, char *line)
         log_error ("error reading from '%s': %s\n", d->file, strerror (errno));
     }
 
-  rc = assuan_send_data (ctx, NULL, 0);
-  if (rc)
-    log_error ("sending data back failed: %s\n", gpg_strerror (rc) );
-
   if (d->is_var)
     ;
   else if (d->is_prog)
     {
-#ifndef HAVE_W32CE_SYSTEM
       if (pclose (fp))
-        log_error ("error running '%s': %s\n", d->file, strerror (errno));
-#endif
+        cancelled = 1;
     }
   else
     fclose (fp);
+
+  rc = assuan_send_data (ctx, NULL, cancelled);
+  if (rc)
+    log_error ("sending data back failed: %s\n", gpg_strerror (rc) );
+
   return 1;
 }
 
